#=======================================================================
#	$Id: UriMap.pm,v 1.1 2005/12/09 10:50:05 pythontech Exp $
#	Map between secions of URI space
#	Copyright (C) 2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::UriMap;
use strict;

sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    die "$class: map not configured\n" unless $self->{'map'};

    $self->{'map'} = {%{$self->{'map'}}}; # Take a copy
    $self->{'rev'} = undef;	# Reverse map
    $self->{'regexp'} = undef;	# Regexp to match all expanded qprefixes
    $self->{'last'} = undef;	# Keep side effect for lastq method.
    return $self;
}

#-----------------------------------------------------------------------
#	Add a mapping
#-----------------------------------------------------------------------
sub add {
    my($self, $fromuri,$touri) = @_;
    my $newto = $self->map($touri);
    $self->{'map'}->{$fromuri} = $newto;
    # Invalidate cached stuff
    $self->{'regexp'} = undef;
    if ($self->{'rev'}) {
	$self->{'rev'}->{'rev'} = undef
	    if $self->{'rev'}->{'rev'} eq $self;
	$self->{'rev'} = undef;
    }
}

#-----------------------------------------------------------------------
#	Given a URI, map to another
#-----------------------------------------------------------------------
sub map {
    my($self, $uri) = @_;
    if (! $self->{'regexp'}) {
	$self->{'regexp'} = &_re($self->{'map'});
    }
    if (my($prefix,$rest) = $uri =~ $self->{'regexp'}) {
	return $self->{'map'}->{$prefix} . $rest;
    }
    return $uri;
}

#-----------------------------------------------------------------------
#	Clone
#-----------------------------------------------------------------------
sub clone {
    my $self = shift;
    my $clone = {
	map => { %{$self->{'map'}} },
	regexp => $self->{'regexp'},
	rev => $self->{'rev'},
    };
    bless $clone, ref($self);
    return $clone;
}

#-----------------------------------------------------------------------
#	Make a reverse map
#-----------------------------------------------------------------------
sub reverse {
    my $self = shift;
    if (! $self->{'rev'}) {
	my %rmap;
	while (my($k,$v) = each %{$self->{'map'}}) {
	    $rmap{$v} = $k;
	}
	my $rev = ref($self)->new(map => \%rmap);
	$self->{'rev'} = $rev;
    }
    return $self->{'rev'};
}

#-----------------------------------------------------------------------
#	For compaibility with QMap
#-----------------------------------------------------------------------
sub decode {
    my($self, $uri) = @_;
    $self->map($uri);
}

sub encode {
    my($self, $uri) = @_;
    $self->reverse->map($uri);
}

#-----------------------------------------------------------------------
#	Given a hash {$fromuri => $touri}, build a regexp matching URIs
#	beginning with any $fromuri.
#	N.B. we want to match longest first, so that e.g. with qmap
#	  http://foo/ -> a:
#	  http://foo/bar/ -> b:
#	then http://foo/bar/baz should be mapped as b:baz
#	rather than a:bar/baz .  It is sufficient to sort the
#	prefixes into reverse order.
#	I don't know how clever Perl's regexp system is - will it
#	amalgamate alternatives to avoid backtracking? For example,
#	/^(abcdef|abc|abd)/ can be implemented as /^ab(c(def)?|d)/
#	(ignoring grouping).  If this turns out to be a hotspot,
#	we may consider making this optimisation manually.
#-----------------------------------------------------------------------
sub _re {
    my $map = shift;
    my $re = ("^(" .
	      join("|",
		   map {quotemeta} 
		   sort {$b cmp $a}
		   keys %$map) .
	      ")(.*)\$");
    return $re;
}

1;
